using System;
using System.Management;
using System.Net;
using System.IO;
using System.Collections;
using System.Text;

namespace ESS
{
    /// <summary>
    /// Gathers information on computers remotely.
    /// </summary>
    class EagleSpreadSheet
    {
        /// <summary>
        /// Set to true if you want to output as CSV, set to fase if you want XML.
        /// </summary>
        public bool Output_Format_CSV;

        /// <summary>
        /// If you want a progress bar updated use this. You must set CheckForIllegalCrossThreadCalls to false.
        /// </summary>
        public System.Windows.Forms.ProgressBar external_progress_bar;

        /// <summary>
        /// If you want a label to tell the user what is happening internally set this value.
        /// </summary>
        public System.Windows.Forms.Label run_status_label;

        /// <summary>
        /// Set this value to use a user account other than current user account.
        /// Should be of the format DOMAIN\Username.
        /// </summary>
        public string userName;

        /// <summary>
        /// Set this value to use a user account other than current user account.
        /// </summary>
        /// <remarks>Stored as public string in memory - no attempt at confidentiality is made.</remarks>
        public string userPass;

        /// <summary>
        /// Stream the program uses to write output.
        /// Although normally this would be private, I wanted the program to be able to retry if
        /// it encountered an error, and I wanted most exception handling to be in a single place.
        /// Because exception handling happens outside this program I didn't have a good place to
        /// put a finally {} clause here, so I've made fout public. Turns out you can call Close()
        /// more than once on a filehandle, so put your [EagleSpreadSheet Instance].fout.Close() in
        /// the finally {} clause in your external class.
        /// </summary>
        public TextWriter fout = null;

        private string output_file;
        private ArrayList aComputerList;

        /// <summary>
        /// This is the generic storage container for all computer-related data gathered by this program.
        /// </summary>
        private struct ComputerInfo
        {
            /// <summary>
            /// Hostname of the remote computer. Obtained using a DNS query, so this may be the FQDN.
            /// </summary>
            public string hostname;
            
            /// <summary>
            /// IP address of the remote computer.Obtained through DNS query.
            /// </summary>
            public string ip;
            
            /// <summary>
            /// MAC address of the remote computer.
            /// </summary>
            public string mac;

            /// <summary>
            /// Domain the remote computer is connected to. Empty for workgroup computers.
            /// </summary>
            public string domain;
            
            /// <summary>
            /// Serial number of the remote computer.
            /// </summary>
            /// <remarks>
            /// This value can be changed in the BIOS on some computers and
            /// not all manufacturers use it...stay safe.
            /// </remarks>
            public string serial;
            
            /// <summary>
            /// Make of the remote computer.
            /// </summary>
            /// <remarks>
            /// Surprises happen with this sometimes so be careful.
            /// </remarks>
            public string make;

            /// <summary>
            /// Model of the remote computer.
            /// </summary>
            /// <remarks>
            /// This isn't reliable - Dell uses a different field in Win32_ComputerSystemProduct than IBM.
            /// </remarks>
            public string model;
            
            /// <summary>
            /// Size of C: in Gigabytes.
            /// </summary>
            public string hdsize;
            
            /// <summary>
            /// Size of memory in Megabytes.
            /// </summary>
            public string memsize;
            
            /// <summary>
            /// Type and speed of the CPU.
            /// </summary>
            /// <remarks>
            /// Often contains a lot of extra whitespace (presumably padding for System Properties dialog).
            /// </remarks>
            public string cputype;
            
            /// <summary>
            /// Operating system of the remote computer.
            /// </summary>
            public string os;
            
            /// <summary>
            /// This is a filehandle to the file you want to output. The idea is it will make
            /// the filehandles a lot easier to manage by putting it all in one spot.
            /// For console applications set to value Console.Out.
            /// </summary>
            public TextWriter OUTPUT;
            
            /// <summary>
            /// This is the WMI scope used to query the remote PC for information. The scope is
            /// stored here to improve performance and make each of the GetXXXInfo() methods simpler
            /// since now they don't have to create their own connection to the remote PC.
            /// </summary>
            public ManagementScope scope;
        }

        private bool bGetIp;
        private bool bGetMac;
        private bool bGetHostname;
        private bool bGetDomain;
        private bool bGetMake;
        private bool bGetModel;
        private bool bGetSerial;
        private bool bGetOs;
        private bool bGetHdd;
        private bool bGetRam;
        private bool bGetCpu;
        private bool bGetWarranty;

        /// <summary>
        /// Constructor to use if you're going to read hostnames from a file.
        /// </summary>
        /// <param name="output_file">CSV or XML file to write data to.</param>
        /// <param name="input_file">Plaintext file to read input from, 
        /// one hostname/IP/FQDN per line.</param>
        public EagleSpreadSheet(string output_file, string input_file)
        {
            this.output_file = output_file;
            this.aComputerList = ReadInputFile(input_file);
        }

        /// <summary>
        /// Constructor to use if you're going to scan a single IP address.
        /// Set scan_single_ip to null if you want to scan the local computer.
        /// </summary>
        /// <param name="output_file">CSV or XML file to write data to.</param>
        /// <param name="scan_single_ip">Set to null to use a hostname.</param>
        /// <param name="hostname">String must be entered, but it won't be used if scan_single_ip is set.</param>
        public EagleSpreadSheet(string output_file, IPAddress scan_single_ip, string hostname)
        {
            this.output_file = output_file;
            this.aComputerList = new ArrayList();
            try
            {
                if (scan_single_ip != null)
                {
                    string szHostname = Dns.GetHostEntry(scan_single_ip).HostName;
                    this.aComputerList.Add(szHostname);
                }
                else
                {
                    this.aComputerList.Add(hostname);
                }
            }
            catch (System.Net.Sockets.SocketException e)
            {
                throw (e);
            }
        }

        /// <summary>
        /// Constructor to use if you're going to scan an IP range.
        /// The IP range is actually created here using network code so call this method asynchronously.
        /// This method makes no internal effort to verify startIP is less than endIP.
        /// </summary>
        /// <param name="output_file">CSV or XML file to write data to.</param>
        /// <param name="startIP">First IP to scan.</param>
        /// <param name="endIP">Last IP to scan.</param>
        public EagleSpreadSheet(string output_file, IPAddress startIP, IPAddress endIP)
        {
            this.aComputerList = new ArrayList();
            this.output_file = output_file;

            ArrayList aIP = new ArrayList();
            IPAddress current = startIP;
            
            aIP.Add(current); // This ensures the first and last are entered since there is no less than operator.
            do
            {
                current = NextIP(current);
                aIP.Add(current);
            } while (!current.Equals(endIP));

            foreach (IPAddress ip in aIP)
            {
                try
                {
                    string hostname = Dns.GetHostEntry(ip).HostName;
                    if (!string.IsNullOrEmpty(hostname))
                    {
                        this.aComputerList.Add(hostname);
                    }
                }
                catch (System.Net.Sockets.SocketException)
                {
                    // do nothing if you can't ping a host
                }
            }
        }

        /// <summary>
        /// Use this constructor when scanning the local machine only.
        /// </summary>
        /// <param name="output_file">CSV or XML file to write data to.</param>
        public EagleSpreadSheet(string output_file)
        {
            this.aComputerList = new ArrayList();
            this.aComputerList.Add(Dns.GetHostName());
            this.output_file = output_file;
        }

        /// <summary>
        /// Use this constructor for console applications.
        /// </summary>
        /// <param name="args">Send this constructor your command-line arguments.</param>
        public EagleSpreadSheet(ref string[] args)
        {
            foreach (string s in args)
            {
                switch (s)
                {
                    case "/xml": Output_Format_CSV = false;
                        break;
                    case "/csv": Output_Format_CSV = false;
                        break;
                    default: // no action
                        break;
                }
            }
        }

        /// <summary>
        /// This method MUST be used by caller.
        /// Each value controls whether that property is found or not.
        /// </summary>
        /// <param name="hostname"></param>
        /// <param name="domain"></param>
        /// <param name="ip"></param>
        /// <param name="mac"></param>
        /// <param name="os"></param>
        /// <param name="cpu"></param>
        /// <param name="hdd"></param>
        /// <param name="ram"></param>
        /// <param name="make"></param>
        /// <param name="model"></param>
        /// <param name="serial"></param>
        /// <param name="warranty"></param>
        public void DefineColumns(bool hostname, bool domain, bool ip, bool mac,
            bool os, bool cpu, bool hdd, bool ram, bool make, bool model,
            bool serial, bool warranty)
        {
            this.bGetCpu = cpu;
            this.bGetDomain = domain;
            this.bGetHdd = hdd;
            this.bGetHostname = hostname;
            this.bGetIp = ip;
            this.bGetMac = mac;
            this.bGetMake = make;
            this.bGetModel = model;
            this.bGetOs = os;
            this.bGetRam = ram;
            this.bGetSerial = serial;
            this.bGetWarranty = warranty;
        }

        /// <summary>
        /// After creating an ESS object use this method to start data collection.
        /// Look through the public variables and see if you want to set any of them before running this method.
        /// </summary>
        public void MasterDataCollect()
        {
            int iReadLines = 0;
            string szInput;

            if (output_file == null)
            {
                fout = Console.Out;
            }
            else
            {
                try
                {
                    fout = new StreamWriter(output_file);
                }
                catch (Exception e)
                {
                    throw (e);
                }
            }

            if (run_status_label != null)
            {
                if (aComputerList.Count > 1)
                {
                    run_status_label.Text = "Collecting data from " + aComputerList.Count.ToString() + " hosts...";
                }
                else
                {
                    run_status_label.Text = "Collecting data...";
                }
            }
            
            if (Output_Format_CSV)
            {
                if (bGetHostname) fout.Write("\"Hostname\",");
                if (bGetIp) fout.Write("\"IP\",");
                if (bGetMac) fout.Write("\"MAC\",");
                if (bGetMake) fout.Write("\"Make\",");
                if (bGetModel) fout.Write("\"Model\",");
                if (bGetSerial) fout.Write("\"SerialNumber\",");
                if (bGetHdd)
                    fout.Write("\"HardDisk\",");
                if (bGetRam) fout.Write("\"RAM\",");
                if (bGetCpu) fout.Write("\"CPU\",");
                if (bGetOs) fout.Write("\"OS\",");
                if (bGetDomain) fout.Write("\"Domain\",");
                if (bGetWarranty) fout.Write("\"Warranty\",");
                fout.WriteLine();
            }
            else
            {
                fout.WriteLine("<?xml version=\"1.0\" encoding=\"UTF-8\"?>");
                fout.WriteLine("<AutoSpreadSheet>");
            }

            foreach (object oLine in aComputerList)
            {
                szInput = (string)oLine;

                ComputerInfo myInfo = new ComputerInfo();
                myInfo.OUTPUT = fout;

                // This performs a lookup on the hostname no matter what was entered. May improve performance.
                try
                {
                    myInfo.hostname = Dns.GetHostEntry(szInput).HostName;
                }
                catch (System.Net.Sockets.SocketException)
                {
                    continue;
                }

                if (string.IsNullOrEmpty(userName))
                {
                    myInfo.scope = new ManagementScope(@"\\" + myInfo.hostname + @"\root\CIMV2");
                }
                else
                {
                    ConnectionOptions oConn = new ConnectionOptions();
                    oConn.Username = userName;
                    oConn.Password = userPass;
                    myInfo.scope = new ManagementScope(@"\\" + myInfo.hostname + @"\root\CIMV2", oConn);
                }

                try
                {
                    // This is how you can set a timeout if desired. I'm not going to use it unless I have to.
                    // myInfo.scope.Options.Timeout = new TimeSpan(0, 0, 2);
                    myInfo.scope.Connect();
                }
                catch (Exception)
                {
                    continue; // skip if connection fails
                }

                if (Output_Format_CSV)
                {
                    PrintPropertiesCsv(ref myInfo);
                }
                else
                {
                    PrintPropertiesXml(ref myInfo);
                }

                if (external_progress_bar != null)
                {
                    iReadLines++;
                    external_progress_bar.Value = (iReadLines * 100) / (aComputerList.Count);
                }
            }

            if (!Output_Format_CSV)
                fout.WriteLine("</AutoSpreadSheet>");

            fout.Close();
        }

        private ArrayList ReadInputFile(string input_file)
        {
            TextReader fin;
            string line;
            ArrayList mylist = new ArrayList();

            try
            {
                if (input_file != null)
                {
                    fin = new StreamReader(input_file);
                    while (!string.IsNullOrEmpty(line = fin.ReadLine()))
                    {
                        mylist.Add(line);
                    }
                    fin.Close();
                }
            }
            catch (Exception e)
            {
                throw (e);
            }

            return (mylist);
        }

        private void PrintPropertiesXml(ref ComputerInfo c)
        {
            if (c.hostname == null)
                return;
            if (c.OUTPUT == null)
                return;

            if (bGetHostname || bGetIp || bGetMac || bGetDomain)
                GetNetworkInfo(ref c);
            if (bGetMake || bGetModel || bGetSerial || bGetWarranty)
                GetHardwareInfo(ref c);
            if (bGetCpu) GetCpuInfo(ref c);
            if (bGetRam) GetRamInfo(ref c);
            if (bGetHdd) GetHdInfo(ref c);
            if (bGetRam) GetRamInfo(ref c);
            if (bGetOs) GetOperatingSystem(ref c);

            c.OUTPUT.WriteLine("  <Computer>");
            if (bGetHostname)
                c.OUTPUT.WriteLine("    <Hostname>{0}</Hostname>", c.hostname);
            if (bGetIp) c.OUTPUT.WriteLine("    <IP>{0}</IP>", c.ip);
            if (bGetMac) c.OUTPUT.WriteLine("    <MAC>{0}</MAC>", c.mac);
            if (bGetMake) c.OUTPUT.WriteLine("    <Make>{0}</Make>", c.make);
            if (bGetModel) c.OUTPUT.WriteLine("    <Model>{0}</Model>", c.model);
            if (bGetSerial) c.OUTPUT.WriteLine("    <SerialNumber>{0}</SerialNumber>", c.serial);
            if (bGetHdd)
                c.OUTPUT.WriteLine("    <HardDisk>{0}</HardDisk>", c.hdsize);
            if (bGetRam) c.OUTPUT.WriteLine("    <RAM>{0}</RAM>", c.memsize);
            if (bGetCpu) c.OUTPUT.WriteLine("    <CPU>{0}</CPU>", c.cputype);
            if (bGetOs) c.OUTPUT.WriteLine("    <OS>{0}</OS>", c.os);
            if (bGetDomain) c.OUTPUT.WriteLine("    <Domain>{0}</Domain>", c.domain);
            if (bGetWarranty)
            {
                if (c.make.ToLower().Contains("dell"))
                {
                    c.OUTPUT.Write("<Warranty>http://supportapj.dell.com/support/topics/topic.aspx/ap/shared/support/my_systems_info/en/details?c=au&l=en&s=gen&ServiceTag={0}</Warranty>", c.serial);
                }
            }
            c.OUTPUT.WriteLine("  </Computer>");
        }

        private void PrintPropertiesCsv(ref ComputerInfo c)
        {
            if (c.hostname == null)
                return;
            if (c.OUTPUT == null)
                return;

            if (bGetHostname || bGetIp || bGetMac || bGetDomain)
                GetNetworkInfo(ref c);
            if (bGetMake || bGetModel || bGetSerial || bGetWarranty)
                GetHardwareInfo(ref c);
            if (bGetCpu) GetCpuInfo(ref c);
            if (bGetRam) GetRamInfo(ref c);
            if (bGetHdd) GetHdInfo(ref c);
            if (bGetRam) GetRamInfo(ref c);
            if (bGetOs) GetOperatingSystem(ref c);

            if (bGetHostname) c.OUTPUT.Write("\"{0}\",", c.hostname);
            if (bGetIp) c.OUTPUT.Write("\"{0}\",", c.ip);
            if (bGetMac) c.OUTPUT.Write("\"{0}\",", c.mac);
            if (bGetMake) c.OUTPUT.Write("\"{0}\",", c.make);
            if (bGetModel) c.OUTPUT.Write("\"{0}\",", c.model);
            if (bGetSerial) c.OUTPUT.Write("\"{0}\",", c.serial);
            if (bGetHdd)
                c.OUTPUT.Write("\"{0}\",", c.hdsize);
            if (bGetRam) c.OUTPUT.Write("\"{0}\",", c.memsize);
            if (bGetCpu) c.OUTPUT.Write("\"{0}\",", c.cputype);
            if (bGetOs) c.OUTPUT.Write("\"{0}\",", c.os);
            if (bGetDomain) c.OUTPUT.Write("\"{0}\",", c.domain);
            if (bGetWarranty)
            {
                if (c.make.ToLower().Contains("dell"))
                {
                    c.OUTPUT.Write("\"http://supportapj.dell.com/support/topics/topic.aspx/ap/shared/support/my_systems_info/en/details?c=au&l=en&s=gen&ServiceTag={0}\",", c.serial);
                }
            }

            c.OUTPUT.WriteLine();
        }

        private void GetHardwareInfo(ref ComputerInfo c)
        {
            ManagementObjectCollection oReturnCollection;
            try
            {
                ObjectQuery oQuery = new ObjectQuery("SELECT IdentifyingNumber,Vendor,Name FROM Win32_ComputerSystemProduct");
                ManagementObjectSearcher oSearcher = new ManagementObjectSearcher(c.scope, oQuery);
                oReturnCollection = oSearcher.Get();
                foreach (ManagementObject oReturn in oReturnCollection)
                {
                    try
                    {
                        c.serial = oReturn["IdentifyingNumber"].ToString();
                    }
                    catch (NullReferenceException) { ; }

                    try
                    {
                        c.make = oReturn["Vendor"].ToString();
                    }
                    catch (NullReferenceException) { ; }

                    try
                    {
                        /* This may give bogus results. Dell uses "Name", IBM uses "Version"...stay safe. */
                        c.model = oReturn["Name"].ToString();
                    }
                    catch (NullReferenceException) { ; }
                }
            }
            catch (System.Management.ManagementException)
            {
                return;
            }
        }

        private void GetNetworkInfo(ref ComputerInfo c)
        {
            ManagementObjectCollection oReturnCollection;

            string DNSHostName;
            // get the DNSHostName in case user provided FQDN.
            if (c.hostname.Contains("."))
            {
                string[] split_hostname = c.hostname.Split('.');
                DNSHostName = split_hostname[0];
            }
            else
            {
                DNSHostName = c.hostname;
            }

            try
            {
                ObjectQuery oQuery = new ObjectQuery(
                    "SELECT MACAddress, DNSDomain FROM Win32_NetworkAdapterConfiguration WHERE DNSHostName=\""
                    + DNSHostName + "\"");
                ManagementObjectSearcher oSearcher = new ManagementObjectSearcher(c.scope, oQuery);
                oReturnCollection = oSearcher.Get();
                foreach (ManagementObject oReturn in oReturnCollection)
                {
                    try
                    {
                        c.ip = Dns.GetHostEntry(c.hostname).AddressList[0].ToString();
                    }
                    catch (System.Net.Sockets.SocketException) { ; }

                    try
                    {
                        c.domain = oReturn["DNSDomain"].ToString();
                    }
                    /* Catches exceptions caused by workgroup computers. */
                    catch (NullReferenceException) { ; }

                    try
                    {
                        c.mac = oReturn["MACAddress"].ToString();
                    }
                    catch (NullReferenceException) { ; }
                }
            }
            catch (System.Management.ManagementException)
            {
                return;
            }
        }

        private void GetHdInfo(ref ComputerInfo c)
        {
            ManagementObjectCollection oReturnCollection;

            try
            {
                ObjectQuery oQuery = new ObjectQuery("SELECT Size FROM Win32_LogicalDisk where Name=\"C:\"");
                ManagementObjectSearcher oSearcher = new ManagementObjectSearcher(c.scope, oQuery);
                oReturnCollection = oSearcher.Get();
                foreach (ManagementObject oReturn in oReturnCollection)
                {
                    try
                    {
                        /* Denominator converts bytes to Gigabytes. */
                        long hdsize = Convert.ToInt64(oReturn["Size"]) / (1024 * 1024 * 1024);
                        c.hdsize = hdsize.ToString() + " GB";
                    }
                    catch (NullReferenceException) { ; }
                }
            }
            catch (System.Management.ManagementException)
            {
                return;
            }
        }

        private void GetCpuInfo(ref ComputerInfo c)
        {
            ManagementObjectCollection oReturnCollection;

            try
            {
                ObjectQuery oQuery = new ObjectQuery("SELECT Name FROM Win32_Processor");
                ManagementObjectSearcher oSearcher = new ManagementObjectSearcher(c.scope, oQuery);
                oReturnCollection = oSearcher.Get();
                foreach (ManagementObject oReturn in oReturnCollection)
                {
                    try
                    {
                        c.cputype = oReturn["Name"].ToString();
                    }
                    catch (NullReferenceException) { ; }
                }
            }
            catch (System.Management.ManagementException)
            {
                return;
            }
        }

        private void GetOperatingSystem(ref ComputerInfo c)
        {
            ManagementObjectCollection oReturnCollection;

            try
            {
                ObjectQuery oQuery = new ObjectQuery("SELECT Caption FROM Win32_OperatingSystem");
                ManagementObjectSearcher oSearcher = new ManagementObjectSearcher(c.scope, oQuery);
                oReturnCollection = oSearcher.Get();
                foreach (ManagementObject oReturn in oReturnCollection)
                {
                    try
                    {
                        c.os = oReturn["Caption"].ToString();
                    }
                    catch (NullReferenceException) { ; }
                }
            }
            catch (System.Management.ManagementException)
            {
                return;
            }
        }

        private void GetRamInfo(ref ComputerInfo c)
        {
            ManagementObjectCollection oReturnCollection;

            try
            {
                ObjectQuery oQuery = new ObjectQuery("SELECT EndingAddress FROM Win32_MemoryArray");
                ManagementObjectSearcher oSearcher = new ManagementObjectSearcher(c.scope, oQuery);
                oReturnCollection = oSearcher.Get();
                foreach (ManagementObject oReturn in oReturnCollection)
                {
                    try
                    {
                        c.memsize = oReturn["EndingAddress"].ToString();
                        Int64 iMem = (int.Parse(c.memsize) + 1) / 1024;
                        c.memsize = iMem + " MB";

                        /* I chose not to use this code because it makes for really weird numbers
                         * when computers use shared memory for the onboard video. */
                        /*if (iMem >= 1024)
                        {
                            c.memsize = (((double)iMem) / 1024.0) + " GB";
                        }
                        else
                        {
                            c.memsize = iMem + " MB";
                        }*/
                    }
                    catch (NullReferenceException) { ; }
                }
            }
            catch (System.Management.ManagementException)
            {
                return;
            }
        }

        /// <summary>
        /// Computes the next valid IP address, so NextIP(192.168.1.255) = 192.168.2.0.
        /// Should return 0.0.0.0 if input is 255.255.255.255.
        /// </summary>
        /// <param name="i">IPAddress object.</param>
        /// <returns>IPAddress object as the value of the next IP address.</returns>
        /// <remarks>This code is only designed and intended to manage IPv4/32-bit IP addresses.</remarks>
        private IPAddress NextIP(IPAddress i)
        {
            byte[] baIP = i.GetAddressBytes();
            int iArrayPosition = 3;
            bool isOverflow = true;
            while ((iArrayPosition >= 0) && (isOverflow))
            {
                if (baIP[iArrayPosition] == byte.MaxValue)
                {
                    baIP[iArrayPosition] = 0;
                    iArrayPosition--;
                    isOverflow = true;
                }
                else
                {
                    baIP[iArrayPosition] += 1;
                    isOverflow = false;
                }
            }
            return(new IPAddress(baIP));
        }
    }
}
